import os
import sys
import threading

import pytest

from strawberry.exceptions import MissingFieldAnnotationError
from strawberry.exceptions.handler import (
    reset_exception_handler,
    setup_exception_handler,
    strawberry_threading_exception_handler,
)


def test_exception_handler(mocker):
    print_mock = mocker.patch("rich.print", autospec=True)

    class Query:
        abc: int

    exception = MissingFieldAnnotationError("abc", Query)

    strawberry_threading_exception_handler(
        (MissingFieldAnnotationError, exception, None, None)
    )

    assert print_mock.call_args == mocker.call(exception)


def test_exception_handler_other_exceptions(mocker):
    print_mock = mocker.patch("rich.print", autospec=True)
    original_exception_mock = mocker.patch(
        "strawberry.exceptions.handler.original_exception_hook", autospec=True
    )

    exception = ValueError("abc")

    strawberry_threading_exception_handler((ValueError, exception, None, None))

    assert print_mock.called is False
    assert original_exception_mock.call_args == mocker.call(ValueError, exception, None)


def test_exception_handler_uses_original_when_rich_is_not_installed(mocker):
    original_exception_mock = mocker.patch(
        "strawberry.exceptions.handler.original_exception_hook", autospec=True
    )

    mocker.patch.dict("sys.modules", {"rich": None})

    class Query:
        abc: int

    exception = MissingFieldAnnotationError("abc", Query)

    strawberry_threading_exception_handler(
        (MissingFieldAnnotationError, exception, None, None)
    )

    assert original_exception_mock.call_args == mocker.call(
        MissingFieldAnnotationError, exception, None
    )


def test_exception_handler_uses_original_when_libcst_is_not_installed(mocker):
    original_exception_mock = mocker.patch(
        "strawberry.exceptions.handler.original_exception_hook", autospec=True
    )

    mocker.patch.dict("sys.modules", {"libcst": None})

    class Query:
        abc: int

    exception = MissingFieldAnnotationError("abc", Query)

    strawberry_threading_exception_handler(
        (MissingFieldAnnotationError, exception, None, None)
    )

    assert original_exception_mock.call_args == mocker.call(
        MissingFieldAnnotationError, exception, None
    )


@pytest.mark.skipif(
    sys.version_info < (3, 8), reason="threading.excepthook is only available in 3.8+"
)
def test_setup_install_handler(mocker):
    reset_exception_handler()
    setup_exception_handler()

    assert threading.excepthook == strawberry_threading_exception_handler


@pytest.mark.skipif(sys.version_info >= (3, 8), reason="test for python < 3.8")
def test_setup_install_handler_does_add_attribute(mocker):
    reset_exception_handler()
    setup_exception_handler()

    assert hasattr(threading, "excepthook") is False


@pytest.mark.skipif(
    sys.version_info < (3, 8), reason="threading.excepthook is only available in 3.8+"
)
def test_setup_does_not_install_handler_when_disabled_via_env(mocker):
    reset_exception_handler()

    mocker.patch.dict(os.environ, {"STRAWBERRY_DISABLE_RICH_ERRORS": "true"})

    setup_exception_handler()

    assert threading.excepthook != strawberry_threading_exception_handler
